; How to write C programs from Scheme...

(provide 'langc)

; defining segments in a C program...

(define C:*HEADER*)	; where to put #include's
(define C:*CPP*)	; where to put #define's
(define C:*DECL*)	; where to declare types, data, functions
(define C:*CODE*)	; where to put code


;------------------------- Registering C objects ------------------------------
(define C::types (stk:new))	; types
(define C::functions (stk:new))	; functions
(define C::data (stk:new))	; first-class data
(define C::wf-macros (stk:new))	; well-founded macros
(define C::hk-macros (stk:new))	; hackish macros

(define C::include (stk:new))	; include files
(define C::extern (stk:new))	; extern declared objects

(define C::current-block (stk:new)) ; block where to define data...
(stk:push! C::current-block C::data)


;--- Remembering information about C objects...
; category: type, function, data, wf-macro, hk-macro?
; is it extern?
; is it being dumped? already dumped?
; function, data: type, possible value
; wf-macro, hk-macro: body


;--- Remembering information about C types...

(define (C::make-type name value)
   (*****))

(define (C:check-type type)
   #f) ;.....

(define (C:typedef def type)
   (C:check-type def)
   (stk:push! C::types def))


(define (find-variable-in-block block name)
   (list-find
      (lambda (x)
	 (and?
	    (eqv? (cadddr x) name)
	    (eqv? (car x) 'datum)))
      (lambda (x r i l) x)
      (lambda (i l) #f)
      (stk->list block)))

(define (C:declare-variable name type)
   (C:check-type type)
   (let ((block (stk:peek C::current-block)))
   (ASSERT (lambda (name) (not (find-variable-in-block block name))) name)
   (stk:push! block (list 'datum #f #f name type))))


(define (C:define-func . x)
   *****)

(define (C:attribute . x)
   *****)

;--- Parsing and Typing C...
; it would be nice to have a program that parses C headers,
; so we can safely manipulate them in a system-independent way...

; and perhaps whole C programs, for the metatranslator project.

;--- Interpreting C programs
; might be funny to define the semantics of C programs in a HLL,
; so we can apply a partial evaluator to them...

;--- Printing C
; This is needed to dump routines or whole C programs...


;---------------------------- Dumping a C program ----------------------------

(define (C:Dump-block post pre . x)
   (map post (list "{" (map pre x) "}")))

(define (C:Dump cno) ; a named C object...
   (if (not (already-dumped cno))
      (case (C:category cno)
         ((type)	(C:Dump-Type cno))
         ((function)	(C:Dump-Func cno))
         ((datum)	(C:Dump-Data cno))
         ((wf-macro)	(C:Dump-WMac cno))
         ((hk-macro)	(C:Dump-HMac cno)))))
		  

(define (C:Dump-Program)
   (with-new-omp
      (set! C:*HEADER* (omp:fork!))	; #include's
      (set! C:*CPP* (omp:fork!))	; #define's
      (set! C:*DECL* (omp:fork!))	; types, data, functions
      (set! C:*CODE* (omp:fork!))	; code
      (map C:RecDump C:extern)))


;--- Dumping to a File
(define C:Filename)
(define (C:Dump-File file)
   (with-output-to-file file (lambda () (C:Dump-Program))))
