;----------------------------- OTOP/vp.scm -----------------------------------
"
   This is the OTOP implementation of the LLL virtual processor ``kernel''
See ~moose/info/OTOP.nfo for more info about this.
"


(require 'otop)
(provide 'vp)


"
Imported words:

declare-variable
	will declare a variable if not already defined
	by, e.g. some optimization module

"

(C:declare-variable 'TOS 'Value_t)


"
------------------------------------------------------------------------------
------------------------- Define Pseudo-ASM machine --------------------------
------------------------------------------------------------------------------


------------------------------- General syntax -------------------------------

define({Terminator},{;})
define({Separator},{})



--------------------------------- Structures ---------------------------------


dnl
dnl ---------------------- C as an assembler -------------------------------
dnl
dnl


(define CELL},	CELLSIZE)		; a cell is 4 bytes long


; ---- abstraction of the Pseudo-ASM instruction set ----

(define Add},		{($1) += ($2) ;})dnl
(define Sub},		{($1) -= ($2) ;})dnl
(define *Mov*},	{($1) = ($2) ;})dnl
(define *Movb*},	{($1) = ($2) ;})dnl
(define *Movw*},	{($1) = ($2) ;})dnl
(define Neg},		{($1) = - ($1) ;})dnl
(define Xchg},		{R0 = $1 ; $1 = $2 ; $2 = R0 ;})dnl
(define },		{})dnl
(define },		{})dnl
dnl
dnl
(define point},	{ifelse({$2},{},{*((cell*)$1)},{((cell*)$1)[$2]})})dnl
dnl
dnl
(define PC},	{esi})dnl	program counter
(define DSP},	{edi})dnl	data stack pointer
(define DSINC},	{CELL})dnl	increment for data stack -- grow up
(define TOS},	{ebx})dnl	buffered top of stack
(define RSP},	{esp})dnl	return stack pointer
(define RSINC},	{-CELL)})dnl	increment for return stack -- grow down

+----(define FRAME},		{ebp})	; a frame pointer for heap manipulation
(define R0},		{eax})	; a scratch register
(define R0L},		{al})	; its lowest byte
(define R0H},		{ah})	; its second byte
(define R1},		{edx})	; second scratch register
(define COUNT},		{ecx})	; count register
(define COUNTLO},	{cl})	; count register -- low byte

(define INSTRTBL},	InstrT)		; table for bytecode threading

(define PopData},	{Mov({$1},TOS)
			 Add(DSP,-DSINC)})
(define PushData},	{Add(DSP,DSINC)
			 Mov(point(DSP),TOS)
			 Mov(TOS,{$1})})
(define PopReturn},	{ifelse({$1},,{Add(RSP,-RSINC)},{pop $1})})
(define PushReturn},	{push $1})
(define },		{})
(define },		{})
(define },		{})
(define },		{})

(define Enter},	{Add(PC,CELL)})
(define Next},	{jmp [PC]})



--------------------------- Pseudo-ASM registers -----------------------------

defdef({PC},	{PC})dnl	-- program counter
defdef({DSP},	{DSP})dnl	-- data stack pointer
defdef({DSINC},	{(-CELL})dnl	-- increment for data stack, grow down
defdef({TOS},	{TOS})dnl	-- buffered top of stack
defdef({TOSL},	{LowWord(TOS)})dnl	-- its lowest word
defdef({TOSLL},	{LowByte(TOS)})dnl	-- its lowest byte
defdef({TOSLH},	{SecondByte(TOS)})dnl	-- its second byte
defdef({RSP},	{RSP})dnl	-- return stack pointer
defdef({RSINC},	{(-CELL)})dnl	-- increment for return stack -- grow down
defdef({FRAME},	{})dnl	-- a frame pointer for heap manipulation
defdef({R0},	{})dnl	-- a scratch register
defdef({R0L},	{})dnl	-- its lowest word
defdef({R0LL},	{})dnl	-- its lowest byte
defdef({R0LH},	{})dnl	-- its second byte
defdef({R1},	{})dnl	-- second scratch register
defdef({COUNT},	{COUNT})dnl	-- count register
defdef({COUNTLO},{LowByte(COUNT)})dnl	-- count register -- low byte


dnl --- Define Pseudo-ASM type-casting ---

define({bytecast},{((Byte)$1)})
define({wordcast},{((Word)$1)})
define({cellcast},{((Cell)$1)})


dnl --- Define Pseudo-ASM pointers ---

define({genpoint},
{ifelse({$3},{},
  {(*(($1*) ($2)))},
  {(*( (($1*) ($2)) + ($3) ))})})

define({point},{genpoint({Cell},$@)})
define({bytepoint},{genpoint({Byte},$@)})
define({wordpoint},{genpoint({Word},$@)})


dnl --- Define Pseudo-ASM Instructions ---

define({Mov},{$1=$2})dnl
define({Add},{$1+=$2})dnl
define({Sub},{$1-=$2})dnl
.......


dnl --- Define Pseudo-ASM structure ---

define({Loop},{for(;COUNT!=0;COUNT--)  BEGIN  $1  END})
define({exit_Loop},{break;})
"

;------ Now include the kernel ------
;   When we're ready, uncomment following line:
;(require 'kernel)
