(provide 'gc)
(require 'ref)


"
THESE STRINGS COMMANDS ARE TO BE DISREGARDED;
THEY'RE A DRAFT ABOUT HOW TO GIVE A *SPECIFICATION* TO THE GC...

(define-signature 'gc-system
   (: state-space (LL-Type))
   (: state (ref-type (state-space)))
   (: U (LL-Type-Universe))
   (: allocator (for-all Type U
	           (| (subtype (: x U) (for-all y)
		      (memory-overflow-exception) )
	    
A memory manager is:
* a universe of low-level types
* a set of roots
* a set of allocators, that generate a *new* variable of given type,
 independent with all others, or generate a memory overflow exception
* a set of operators on roots
* a set of memory access operators to/from roots

Additional features:
* commutativity: proofs that the above mostly commute,
 whenever accessed variables are independent
* a set of generating
* liveliness:
 
* compactness:
 there exists a (small) constant c1 such
 space can be allocated if the set of live objects is small enough
 so that
 c1*(total size of live objects+total size of objects to allocate)
 < total size of available memory
* realtime:
 execution time for the above primitive is bounded by
 a (small) constant time the number memory accesses needed.
* weak pointers and destructors:
 some pointers can be declared as "weak", with a continuation being
 activated when no more "strong" pointers point to the same object.
"


"
define gc-mark
   (params:
      signature
        // A machine with registers
	C: CPU

	// A pool of memory
	M: Memory

      	// A universe of (low-level) types
        U: Type-Universe;

	// The type for values in registers
	Atom: Type
	Pointer: U->Type
	V = Atom | Pointer (T:U)

	// accessors that allows U-typed communication from M to C
        accessors: ...

	// mutators that allows U-typed modifications from C to M
	mutators: ...

	// A list of free objects cells
	L: ...

	// An operator to create a new object if enough memory is available
        raw-alloc: forall (T:U) new-location(T)

      end
   )
=
{
   uses params;

   let GC-Type-Wrap T =
      relocated-to (Pointer T)
	
   let alloc

   export alloc = 
}
"


; A dumb machine
(define (inline) 'unspecified)

(define hw::size 10000)
(define hw::mem (make-vector hw::size 'uninitialized))
(define (n->v n) n)
(define (hw:get p)
   (if (zero? (remainder p 4))
       (vector-ref hw::mem (quotient p 4))
       (error "Alignment error reading memory at " p)))
(define (hw:set! p v)
   (if (zero? (remainder p 4))
       (vector-set! hw::mem (quotient p 4) v)
       (error "Alignment error writing memory at " p)))

(define number-of-registers 32) ; registers & global variables
(define register-base 100)
(define (register:get reg)
   (vector-ref hw::mem (+ register-base reg))
(define (register:set! reg v)
   (vector-set! hw::mem (+ register-base reg) v))

(register:set! -1 (+ 1 (* 4 number-of-registers)))
(register:set! -2 (+ 1 (* 4 register-base)))

(define heap-base (* 4 (+ number-of-registers register-base))
(define heap-size (- hw::size heap-base))


; "Real" code begins below...

; Notes:
;
; assumes the following "types":
; "v" indicates a generic value, pointer, offset or integer,
;     as found in the FORTH stack.
; "i" indicates a value-encoded (signed) system integer.
; "u" indicates a value-encoded unsigned system integer.
; "n" indicates a high-level number with arbitrary many digits.
; "o" indicates a value-encoded offset suitable for addition to pointers
; "p" indicates a value-encoded pointer
;
; The following definitions build some system abstractions
; out of a two-complement binary machine
; assuming the following BCPL-style primitives:
; "v+" adds values
; "v-" substracts values
; "v=?" compares values
; "vu<=?", "vs<=?" to compare two values unsigned or signed
; "v-bitwise-and" bit-wise and on values
; "v-bitwise-or" bit-wise or on values
; "v-shl"
; "v-shr"
; "v-sar"
; "n->v" converts a high-level number into system value
; "v->n" converts a value into high-level number
; "hw:get" gets a value from the memory hardware
; "hw:set!" sets a value into memory hardware

(define v0
   "value 0"
   (inline)
   (n->v 0))
(define v1
   "value 1"
   (inline)
   (n->v 1))
(define TYPEBITS
   "Number of bits sacrificed to pointer/integer differentiation"
   (inline)
   v1)
(define TYPEMASK
   (inline)
   (v- (v-shl v1 TYPEBITS) v1))
(define VALUEMASK
   (inline)
   (v-bitwise-not TYPEMASK))
(define TYPE-INT
   "Value of the type bits for an integer"
   (inline)
   v0)
(define TYPE-PTR
   "Value of the type bits for a pointer"
   (inline)
   v1)
(define (v->TYPE v)
   (inline)
   (v-bitwise-and v TYPEMASK))
(define vNULL
   "NULL pointer value"
   (inline)
   TYPE-PTR)
(define (v->i v)
   "transforms value into integer"
   (inline)
   (v-shl v TYPEBITS))
(define (cast-to-i v)
   "casts a value into integer"
   (inline)
   (v-bitwise-and v VALUEMASK))
(define (cast-to-p v)
   "casts a value into pointer"
   (inline)
   (v-bitwise-or v TYPE-PTR))
(define (p? v)
   (inline)
   (v=? (v->TYPE v) TYPE-PTR))
(define CELLABITS
   "Number of bits to shift an address"
   "to get the number of cells since NULL pointer"
   (inline)
   (n->v 2))
(define (i->o v)
   "converts a number of cells into an offset to add to pointers"
   "like CELLS does in ANS Forth"
   (inline)
   (v-shl v (v- CELLABITS TYPEBITS)))
(define (n->i n)
   (inline)
   (v->i (n->v n)))
(define (n->o n)
   "converts a high-level number of cells into an offset to add to pointers"
   (inline)
   (i->o (n->i v)))
(define (p->addr)
   (inline)
   (v+ p (v-neg TYPE-PTR)))
(define (mem:get p)
   (inline)
   (hw:get (p->addr p)))
(define (mem:set! p v)
   (inline)
   (hw:set! (p->addr p) v))
(define uninitialized-value
   "value to use when pseudo-initializing an object"
   'uninitialized)

; type information should be added to all the value munging
; for easier program specification and/or partial evaluation,
; by adding a proper type declaring/checking layer over values...
(define po+ v+)
(define oo+ v+)
(define pp- v-)
(define oo- v-)


; Should be rewritten later in a much more modular style,
; when a module system is available:
; we should have a recursive combination of higher-order functors
; with mutual dependencies (ouch)

;(define (stupid-flat-mark-compact params))

; Assumes the following words:
;   heap-base	"base of the heap: a pointer value"
;   heap-size	"size of the heap: an offset/size value"

; Convention:
; foo:bar means a public member of module foo named bar
; foo::zap means a private (unexported) member of same module named zap
  

;;; Overflow behavior
(define (space-overflow)
  (throw 'memory-overflow))

;;; These two re-export the former arguments...
(define gc::heap-base
  "base of the heap"
  heap-base)

(define gc::heap-size
  "size of the heap"
  heap-size)

(define gc::heap-end
  "end of the heap"
  (po+ heap-base heap-size))

;;; Global heap state
(define gc::heap-bottom
  "Bottom of free space"
  (ref:new gc:heap-base))

(define gc::heap-top
  "top of free space"
  (ref:new gc:heap-end))

(define (gc::free-size)
  (pp- (ref:get gc::heap-top) (ref:get gc::heap-bottom)))

(define (gc::used-size)
  (oo- gc::heap-size (gc::free-size)))

(define (gc::enough-space size)
  (o<= size (gc::free-size)))

;;; Low level allocation routines
(define (gc:llar:reserve-space-or-action size action)
  (if (not (gc::enough-space size))
      (action)))

(define (gc:llar:get-space-high size)
  "get reserved space from top of free space"
  (ref:apply! gc::heap-top po- size)
  (ref:get gc::heap-top))

(define (gc:llar:get-space-low size)
  "get reserved space from bottom of free space"
  (ref:postapply! gc::heap-bottom po+ size))

;;; Simple Stupid GC behavior over the former:
; trigger more and more full GCs until there is enough free space,
; or the GC was unsuccessful at saving any space, in which case
; a failure is signaled.
; * for GCs that already do full tracing in one pass,
;  another primitive is provided.
; * failure might be throwing an exception out of reserved space...

(define (gc::reserve-space-or-one-gc size)
  (reserve-space-or-action
     size
     (lambda ()
       (gc:full-gc)
       (if (not (gc::enough-space size))
	   (space-overflow)))))

(define (gc::reserve-space-or-multiple-gc size)
  (reserve-space-or-action
     size
     (lambda ()
       (let*
	   ((previous-used-size (gc::used-size))
	    (gc-dummy-result (gc:full-gc))
	    (new-used-size (gc::used-size)))
	 (if (o<= previous-used-size new-used-size)
	     (space-overflow)
	     (reserve-space-or-multiple-gc size))))))

;;; Accessing objects in GC'ed memory
(define (gc::get zone offset)
  "get value at given offset in given zone"
  (mem:get (gc:enc:field zone offset)))

(define (gc::set! zone offset value)
  "set value at given offset in given zone"
  (mem:set! (gc:enc:field zone offset) value))

;;; The same, with bound checking
(define (gc:get zone offset)
  "get value at given offset in given zone"
  (ASSERT (lambda (x) (< -1 (cdr x) (gc::zone-size (car x))))
	  (cons zone offset)
	  "Trying to read out of bound")
  (gc::get zone offset))

(define (gc:set! zone offset value)
  (ASSERT (lambda (x) (< -1 (cdr x) (gc::zone-size (car x))))
	  (cons zone offset)
	  "Trying to write out of bound")
  (gc::set! zone offset value))

;;; Encoding of GC information
; Attributes of objects:
; size; type: raw or cooked; marked or not; forward pointer
; one cell for size, one cell for mark&sweep forward pointer

;;; Size/Type field
(define gc::attr-size
  (n->o 1))

(define (gc::code-attr-size attr size)
  (v-bitwise-or size attr))

(define (gc::decode-size v)
  (v-bitwise-and v VALUEMASK))

(define (gc::decode-attr v)
  (v-bitwise-and v TYPEMASK))

(define gc::attr-raw
  "tag for raw pointer-less memory"
  TYPE-INT)

(define gc::attr-cooked
  "tag for cooked pointer-less memory"
  TYPE-PTR)

(define (gc::field zone offset)
  (po+ zone offset))

(define gc::attr-field
  (n->o -1))

(define (gc::attr:set! zone attr size)
  (gc::set! zone gc:attr-field (gc::code-attr-size attr size)))

(define (gc::attr:get zone)
  (gc::get zone gc:attr-field))

(define (gc::get-size zone)
  (gc::decode-size (gc::attr:get zone)))


;;; Forward/Mark field
(define gc::mark-field
  (n->o -2))	; only needed by mark&sweep algorithm.

(define (gc::get-mark zone)
  (gc::get zone gc::mark-field))

(define (gc::set-mark! zone v)
  (gc::set! zone gc::mark-field v))

(define (gc::get-forw zone)
  ((cast-to-p (gc::get-mark zone))))

(define (gc::set-forw! zone forw)
  ((gc::set-mark! (cast-to-i forw)))) ; also clears the mark...

(define (gc::marked? zone)
  (p? (gc::get-mark zone)))

(define (gc::mark! zone)
  (gc::set-mark! zone (cast-to-p (gc::get-mark zone))))

;;; Allocation
(define (gc::init-zone zone attr size)
  ;should we initialize it to zero ?
  (do ((i 0 (+ i 1))) (< i size) (gc::set! zone i unitialized-value))
  (gc::attr:set! zone attr size)
  (gc::unmark-it zone)))

(define (gc:do-alloc attr size)
  (gc::get-space-high gc::hidden-size)
  (let*
      ((header (gc::get-space-low (gc::header-size)))
       (zone (gc::get-space-low size)))
    (gc::init-zone zone attr size)))

(define (gc:alloc attr size)
  (inline)
  (reserve-space (oo+ size gc::size-overhead))
  (gc::do-alloc attr size))

;;; Size information for the allocator
(define gc::hidden-size
  (n->o 1))

(define gc::header-size
  (oo+ gc::attr-size gc::forw-size))

(define gc::size-overhead
  (oo+ gc::header-size gc::hidden-size))

;(define gc::wrap-size
;   (lambda (size)
;      (oo+ size gc::header-size)))

;;; Using high-reserved space as a stack/queue for the tracing GC...
; LIFO means depth-first tracing; FIFO means breadth-first tracing
; while both have same worst-case depth behavior,
; LIFO generally consumes much less depth,
; but requires the current traced offset to be remembered for all objects.
; Hence, FIFO is generally preferred,
; but if computing the base zone for an infix pointer is very cheap,
; for instance when all objects are of the same power-of-two size
; in which case we an infix pointer is used for both base and offset.
(define gc::trace-stack-pointer (ref:new)) ; optimize: use heap-top instead
(define gc::trace-unstack-pointer (ref:new)) ; for FIFO
(define (gc::push! v)
   (ref:apply gc::trace-stack-pointer v+ (n->o -1))
   (mem:set! gc::trace-stack-pointer v))
(define (gc::pop!)
   (let ((v (mem:get gc::trace-stack-pointer)))
     (ref:apply gc::trace-stack-pointer v+ (n->o 1))
     v))
(define gc::queue! gc::push!)
(define (gc::unqueue!)
   (let ((v (mem:get gc::trace-unstack-pointer)))
     (ref:apply gc::trace-unstack-pointer v+ (n->o -1))
     v))
(define (gc::reset-stack)
   (ref:set! gc::trace-stack-pointer gc::heap-top)
   (ref:set! gc::trace-unstack-pointer gc::heap-top))
(define (gc::empty-queue?)
   (eqv? gc::trace-stack-pointer gc::trace-unstack-pointer))

;;; The Tracing algorithm

(define (for-each-cooked-value-in zone action)
   (if (p? (gc::attr:get zone))
       (do ((size (gc::get-size zone))
	    (i 0 (+ i 1)))
	   (< i size)
	 (action (gc::field zone i)))))

(define (for-each-reachable-pointer action)
  ;(save-registers-in-root-structure)
  (gc::push! root-structure)
  (while (not (gc::empty-queue?))
	 (let ((zone (gc::unqueue!)))
	   (for-each-cooked-value-in zone
	      (lambda (p)
		 (let ((v (mem:get p)))
		   (if (p? v)
		       (begin
			 (if (not (gc::marked? v))
			     (begin
			       (gc::mark! v)
			       (gc::queue! v)))
			 (action p))))))))
  ;(restore-registers-in-root-structure)
  (ref:set! gc::heap-top gc::trace-stack-pointer))


;;; Traversing all allocation zones

(define first-zone
  (po+ gc::heap-bottom gc::header-size))

(define (next-zone zone)
  (po+ (po+ zone (gc::get-size zone)) gc::header-size))

(define (for-each-zone action)
  (do ((zone first-zone (next-zone zone)))
      (vu<= zone gc::heap-bottom)
    (action zone)))


;;; Now for the Mark&Compact GC itself

(define (gc:full-gc)
  (gc::clear-all-marks)
  (gc::mark-all-reachable-zones)
  (gc::compute-new-addresses)
  (gc::update-all-pointers)
  (gc::move-zones))


;;; (I) Initialization phase
; It is done incrementally
;Not necessary: done incrementally by allocation and move routines...

(define (gc::clear-all-marks)
  ;(for-each-zone unmark-it)
  #t)

(define (gc::unmark-it zone)
  (gc::set-mark! zone v0))


;;; (II) Mark phase
; Mark phase
(define (mark-pointee-as-reachable p)
  #t)
; already done by the traversal algorithm...
;  (let ((zone (mem:get p)))
;    (gc::set-mark! zone vNULL)))

(define (gc::mark-all-reachable-zones)
  (for-each-reachable-pointer mark-pointee-as-reachable))


;;; (III) Virtual compact phase

(define gc::compaction-pointer (ref:new))

(define (compute-its-new-address zone)
  (if (gc::marked? zone)
      (let ((new-zone (po+ gc::compaction-pointer gc::header-size)))
        (gc::set-forw! zone new-zone)
	(ref:apply! gc::compaction-pointer po+ (gc::get-size zone)))))

(define (gc::compute-new-addresses)
  (ref:set! gc::compaction-pointer gc::heap-base)
  (for-each-zone compute-its-new-address))

  
;;; (IV) Update phase
(define (update-it p)
  (mem:set! p (gc::get-forw (mem:get p))))

(define (gc::update-all-pointers)
  (for-each-reachable-pointer update-it)


;;; (V) Actual compact phase
(define (move-it-to-its-new-location zone)
  (let ((new-location (gc::get-forw zone)))
    (if (not (v=? new-location zone))
        (begin
	  (gc::unmark-it zone)
	  (do ((i -1 (1+ i)) (size (gc::get-size zone)))
	      (< i size))
	      (gc::set! new-location i (gc::get zone i))))))

(define (gc::move-zones)
  (ref:set! gc::compaction-pointer gc::heap-base)
  (for-each-zone move-it-to-its-new-location)
  (ref:set! gc::heap-bottom gc::compaction-pointer))

;Charles fiterman:
;Our best luck
;has been with gcAttemptCollection( userIdleDetectFunction ); That
;is calling for a collection but trashing it if interaction is
;Required. In real world applications this works very well.
